from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np

from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from unittest import TestCase

from ..scm import SetCoveringMachineClassifier


class SklearnCompatibilityTests(TestCase):
    def test_sklearn_grid_search(self):
        """
        Test compatibility with sklearn GridSearchCV function

        """
        rnd = np.random.RandomState(0)
        X = np.random.rand(100, 100)
        y = np.random.randint(2, size=100)

        scm_param_grid = {
            "p": [0.01, 0.1, 1, 10, 100],
            "max_rules": [1, 2, 3, 4, 5],
            "model_type": ["conjunction", "disjunction"],
            "random_state": [rnd],
        }
        gscv = GridSearchCV(
            SetCoveringMachineClassifier(), scm_param_grid, n_jobs=1, cv=5
        )
        try:
            gscv.fit(X, y)
        except Exception as e:
            self.fail("GridSearchCV.fit() raised " + e.message + " unexpectedly!")

        gscv = GridSearchCV(
            SetCoveringMachineClassifier(), scm_param_grid, n_jobs=2, cv=5
        )
        try:
            gscv.fit(X, y)
        except Exception as e:
            self.fail(
                "GridSearchCV.fit() raised "
                + e.message
                + " unexpectedly when running with 2 jobs!"
            )

    def test_sklearn_pipeline(self):
        """
        Test compatibility with sklearn pipelines

        """
        rnd = np.random.RandomState(0)
        X = np.random.rand(10, 10)
        y = np.random.randint(2, size=10)

        scm = SetCoveringMachineClassifier(
            model_type="conjunction", p=0, max_rules=3, random_state=rnd
        )
        pipeline = Pipeline([("Scale", StandardScaler()), ("SCM", scm)])
        try:
            pipeline.fit(X, y)
        except Exception as e:
            self.fail("Pipeline.fit() raised " + e.message + " unexpectedly!")



from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import sys

from numpy import infty as inf
from unittest import TestCase
from sklearn.utils import estimator_checks

from .._scm_utility import find_max


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class UtilityTests(TestCase):
    def setUp(self):
        """
        Called before each test

        """
        pass

    def tearDown(self):
        """
        Called after each test

        """
        pass

    def test_1(self):
        """
        Dummy test #1
        """
        X = np.array([[1, 2, 2, 2, 3, 4]], dtype=np.double).reshape(-1, 1).copy()
        y = np.array([0, 1, 0, 1, 1, 1])
        p = 1
        Xas = np.argsort(X, axis=0).T.copy()
        (
            best_utility,
            best_feat_idx,
            best_thresholds,
            best_kinds,
            best_N,
            best_P_bar,
        ) = find_max(p, X, y, Xas, np.arange(X.shape[0]), np.ones(1))
        np.testing.assert_almost_equal(actual=best_utility, desired=1.0)
        np.testing.assert_almost_equal(actual=best_feat_idx, desired=[0])
        np.testing.assert_almost_equal(actual=best_thresholds, desired=[1])
        np.testing.assert_almost_equal(actual=best_kinds, desired=[0])

    def test_2(self):
        """
        Test that hyperparameter p works
        """
        X = np.array([[1, 2, 2, 2, 3, 4]], dtype=np.double).reshape(-1, 1).copy()
        y = np.array([0, 1, 0, 1, 1, 1])
        Xas = np.argsort(X, axis=0).T.copy()
        p = 0.5
        (
            best_utility,
            best_feat_idx,
            best_thresholds,
            best_kinds,
            best_N,
            best_P_bar,
        ) = find_max(p, X, y, Xas, np.arange(X.shape[0]), np.ones(1))

        np.testing.assert_almost_equal(actual=best_utility, desired=1.0)
        np.testing.assert_almost_equal(actual=best_feat_idx, desired=[0, 0])
        np.testing.assert_almost_equal(actual=best_thresholds, desired=[1, 2])
        np.testing.assert_almost_equal(actual=best_kinds, desired=[0, 0])

    def test_3(self):
        """
        Test that feature_weights works
        """
        X = np.array([[1, 1], [1, 0]], dtype=np.double)
        y = np.array([0, 1])
        Xas = np.argsort(X, axis=0).T.copy()
        p = 1.0

        # Equal weights, feat 1 should be the best with utility 1
        (
            best_utility,
            best_feat_idx,
            best_thresholds,
            best_kinds,
            best_N,
            best_P_bar,
        ) = find_max(p, X, y, Xas, np.arange(X.shape[0]), np.ones(X.shape[1]))
        np.testing.assert_almost_equal(actual=best_utility, desired=1)
        np.testing.assert_almost_equal(actual=best_feat_idx, desired=[1])

        # Double weight for feat 1, should be the best with utility 2
        (
            best_utility,
            best_feat_idx,
            best_thresholds,
            best_kinds,
            best_N,
            best_P_bar,
        ) = find_max(p, X, y, Xas, np.arange(X.shape[0]), np.array([1.0, 2.0]))
        np.testing.assert_almost_equal(actual=best_utility, desired=2)
        np.testing.assert_almost_equal(actual=best_feat_idx, desired=[1])

        # 10 times the weight for feat 1, should be the best with utility 10
        (
            best_utility,
            best_feat_idx,
            best_thresholds,
            best_kinds,
            best_N,
            best_P_bar,
        ) = find_max(p, X, y, Xas, np.arange(X.shape[0]), np.array([1.0, 10.0]))
        np.testing.assert_almost_equal(actual=best_utility, desired=10)
        np.testing.assert_almost_equal(actual=best_feat_idx, desired=[1])

    def test_4(self):
        """
        Test that example_idx works
        """
        X = np.array([[1, 1], [0, 0], [1, 0]], dtype=np.double)
        y = np.array([0, 1, 1])
        Xas = np.argsort(X, axis=0).T.copy()
        p = 1.0

        # If example 3 is included, the best feature is feat1
        (
            best_utility,
            best_feat_idx,
            best_thresholds,
            best_kinds,
            best_N,
            best_P_bar,
        ) = find_max(p, X, y, Xas, np.arange(X.shape[0]), np.ones(X.shape[1]))
        np.testing.assert_almost_equal(actual=best_feat_idx, desired=[1])

        # If example 3 is included, the best feature is feat1
        (
            best_utility,
            best_feat_idx,
            best_thresholds,
            best_kinds,
            best_N,
            best_P_bar,
        ) = find_max(p, X, y, Xas, np.array([1, 2], dtype=int), np.ones(X.shape[1]))
        np.testing.assert_almost_equal(actual=best_feat_idx, desired=[0, 1])

    def test_5(self):
        """
        Test that solver return accurate equivalent rules
        """
        X = np.array(
            [
                [1, 1, 0.5, 1],
                [2, 1, 0.5, 1],
                [2, 1, 0.5, 1],
                [3, 1, 1.7, 0],
                [4, 1, 1.7, 0],
                [5, 1, 1.7, 0],
                [6, 1, 1.7, 0],
                [7, 1, 1.7, 0],
            ],
            dtype=np.double,
        )
        y = np.array([0, 0, 0, 1, 1, 1, 1, 1])
        Xas = np.argsort(X, axis=0).T.copy()
        p = 1.0

        (
            best_utility,
            best_feat_idx,
            best_thresholds,
            best_kinds,
            best_N,
            best_P_bar,
        ) = find_max(p, X, y, Xas, np.arange(X.shape[0]), np.ones(X.shape[1]))
        np.testing.assert_almost_equal(actual=best_utility, desired=3.0)
        np.testing.assert_almost_equal(actual=best_feat_idx, desired=[0, 2, 3])
        np.testing.assert_almost_equal(actual=best_thresholds, desired=[2.0, 0.5, 0.0])
        np.testing.assert_almost_equal(actual=best_kinds, desired=[0, 0, 1])

    def test_6(self):
        """
        Test that solver return accurate N and P_bar
        """
        X = np.array(
            [
                [0, 0.5, 0],
                [0, 1.7, 0],
                [1, 0.5, 0],
                [0, 1.7, 0],
                [1, 0.5, 0],
                [1, 1.7, 0],
                [1, 1.7, 0],
                [1, 1.7, 0],
            ],
            dtype=np.double,
        )
        y = np.array([0, 0, 0, 1, 1, 1, 1, 1])
        Xas = np.argsort(X, axis=0).T.copy()
        p = 1.0

        (
            best_utility,
            best_feat_idx,
            best_thresholds,
            best_kinds,
            best_N,
            best_P_bar,
        ) = find_max(p, X, y, Xas, np.arange(X.shape[0]), np.ones(X.shape[1]))
        np.testing.assert_almost_equal(actual=best_N, desired=[2, 2])
        np.testing.assert_almost_equal(actual=best_P_bar, desired=[1, 1])

    def test_random_data(self):
        """
        Random testing
        """
        n_tests = 10  # 10000

        # Using rounding generates cases with equal feature values for examples
        for n_decimals in range(3):

            # The more examples, the more likely we are to have equal feature values
            for n_examples in [10, 100, 1000]:

                # Do this a few times for each configuration
                for _ in range(n_tests):
                    p = max(0, np.random.rand() * 100.0)
                    x = (
                        (np.random.rand(n_examples) * 5.0)
                        .round(n_decimals)
                        .reshape(-1, 1)
                        .copy()
                    )
                    xas = np.argsort(x, axis=0).T.copy()
                    y = np.random.randint(0, 2, n_examples)
                    thresholds = np.unique(x)

                    # Use the solver to find the solution
                    (
                        solver_best_utility,
                        solver_best_feat_idx,
                        solver_best_thresholds,
                        solver_best_kinds,
                        solver_best_N,
                        solver_best_P_bar,
                    ) = find_max(p, x, y, xas, np.arange(n_examples))

                    # Less equal rule utilities
                    le_rule_utilities = []
                    for t in thresholds:
                        rule_classifications = (x <= t).reshape(
                            -1,
                        )
                        N = (~rule_classifications[y == 0]).sum()
                        P_bar = (~rule_classifications[y == 1]).sum()
                        le_rule_utilities.append(N - p * P_bar)

                    # Greater rule utilities
                    g_rule_utilities = []
                    for t in thresholds:
                        rule_classifications = (x > t).reshape(
                            -1,
                        )
                        N = 1.0 * (~rule_classifications[y == 0]).sum()
                        P_bar = 1.0 * (~rule_classifications[y == 1]).sum()
                        g_rule_utilities.append(N - p * P_bar)

                    np.testing.assert_almost_equal(
                        actual=solver_best_utility,
                        desired=max(max(le_rule_utilities), max(g_rule_utilities)),
                    )



"""
    pyscm -- The Set Covering Machine in Python
    Copyright (C) 2017 Alexandre Drouin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
import numpy
from platform import system as get_os_name
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext as _build_ext

# Configure the compiler based on the OS
if get_os_name().lower() == "darwin":
    os_compile_flags = ["-mmacosx-version-min=10.9"]
else:
    os_compile_flags = []


solver_module = Extension(
    "pyscm._scm_utility",
    language="c++",
    sources=["cpp_extensions/utility_python_bindings.cpp", "cpp_extensions/solver.cpp"],
    extra_compile_args=["-std=c++0x"] + os_compile_flags,
    include_dirs=[numpy.get_include()]
)

dependencies = ["numpy<2", "scikit-learn", "six"]

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="pyscm-ml",
    version="1.1.2",
    packages=find_packages(),
    install_requires=dependencies,
    author="Alexandre Drouin",
    author_email="aldro61@gmail.com",
    maintainer="Alexandre Drouin",
    maintainer_email="aldro61@gmail.com",
    description="The Set Covering Machine algorithm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPL-3",
    keywords="machine-learning classification set-covering-machine rule-based-models",
    url="https://github.com/aldro61/pyscm",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    ext_modules=[solver_module],
)
